@@ -1,19 +1,16 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from __future__ import unicode_literals
-from decimal import Decimal
-
+from collections import defaultdict
 from trytond.pool import PoolMeta, Pool
 from trytond.model import ModelView, Workflow, fields
 from trytond.pyson import Eval, If, Bool
 from trytond.transaction import Transaction
 from trytond.tools import grouped_slice
 
-__all__ = ['Invoice', 'InvoiceLine']
+from trytond.modules.product import round_price
 
 
-class Invoice:
-    __metaclass__ = PoolMeta
+class Invoice(metaclass=PoolMeta):
     __name__ = 'account.invoice'
     agent = fields.Many2One('commission.agent', 'Commission Agent',
         domain=[
@@ -24,16 +21,19 @@ class Invoice:
             'invisible': Eval('type') == 'in',
             'readonly': Eval('state', '') != 'draft',
             },
-        depends=['type', 'company', 'state'])
+        depends=['type', 'company', 'state'],
+        help="The agent who receives a commission for the invoice.")
 
     @classmethod
-    @ModelView.button
-    @Workflow.transition('posted')
-    def post(cls, invoices):
+    def _journal_types(cls, invoice_type):
+        return super()._journal_types(invoice_type) + ['commission']
+
+    @classmethod
+    def _post(cls, invoices):
         # Create commission only the first time the invoice is posted
         to_commission = [i for i in invoices
             if i.state not in ['posted', 'paid']]
-        super(Invoice, cls).post(invoices)
+        super()._post(invoices)
         cls.create_commissions(to_commission)
 
     @classmethod
@@ -61,19 +61,27 @@ class Invoice:
 
         super(Invoice, cls).paid(invoices)
 
+        date2commissions = defaultdict(list)
         for sub_invoices in grouped_slice(invoices):
             ids = [i.id for i in sub_invoices]
-            commissions = Commission.search([
-                    ('date', '=', None),
-                    ('origin.invoice', 'in', ids, 'account.invoice.line'),
-                    ])
-            Commission.write(commissions, {
-                    'date': today,
-                    })
+            for commission in Commission.search([
+                        ('date', '=', None),
+                        ('origin.invoice', 'in', ids, 'account.invoice.line'),
+                        ]):
+                date = commission.origin.invoice.reconciled or today
+                date2commissions[date].append(commission)
+        to_write = []
+        for date, commissions in date2commissions.items():
+            to_write.append(commissions)
+            to_write.append({
+                'date': date,
+                })
+        if to_write:
+            Commission.write(*to_write)
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     def cancel(cls, invoices):
         pool = Pool()
         Commission = pool.get('commission')
@@ -99,14 +107,12 @@ class Invoice:
         Commission.delete(to_delete)
         Commission.save(to_save)
 
-    def _credit(self):
-        credit = super(Invoice, self)._credit()
-        credit.agent = self.agent
-        return credit
+    def _credit(self, **values):
+        values.setdefault('agent', self.agent)
+        return super()._credit(**values)
 
 
-class InvoiceLine:
-    __metaclass__ = PoolMeta
+class InvoiceLine(metaclass=PoolMeta):
     __name__ = 'account.invoice.line'
     principal = fields.Many2One('commission.agent', 'Commission Principal',
         domain=[
@@ -118,7 +124,8 @@ class InvoiceLine:
             'invisible': If(Bool(Eval('_parent_invoice')),
                 Eval('_parent_invoice', {}).get('type') == 'in',
                 Eval('invoice_type') == 'in'),
-            }, depends=['invoice_type', 'company'])
+            }, depends=['invoice_type', 'company'],
+        help="The principal who pays a commission for the invoice line.")
     commissions = fields.One2Many('commission', 'origin', 'Commissions',
         readonly=True,
         states={
@@ -159,15 +166,17 @@ class InvoiceLine:
                     self.amount, agent.currency, round=False)
             amount = self._get_commission_amount(amount, plan)
             if amount:
-                digits = Commission.amount.digits
-                amount = amount.quantize(Decimal(str(10.0 ** -digits[1])))
+                amount = round_price(amount)
             if not amount:
                 continue
 
             commission = Commission()
             commission.origin = self
             if plan.commission_method == 'posting':
-                commission.date = today
+                commission.date = self.invoice.invoice_date or today
+            elif (plan.commission_method == 'payment'
+                    and self.invoice.state == 'paid'):
+                commission.date = self.invoice.reconciled or today
             commission.agent = agent
             commission.product = plan.commission_product
             commission.amount = amount
@@ -198,6 +207,30 @@ class InvoiceLine:
     def copy(cls, lines, default=None):
         if default is None:
             default = {}
+        else:
+            default = default.copy()
         default.setdefault('commissions', None)
         default.setdefault('from_commissions', None)
         return super(InvoiceLine, cls).copy(lines, default=default)
+
+
+class CreditInvoiceStart(metaclass=PoolMeta):
+    __name__ = 'account.invoice.credit.start'
+    with_agent = fields.Boolean(
+        "With Agent",
+        help="Check to keep the original invoice's agent.")
+
+    @classmethod
+    def default_with_agent(cls):
+        return True
+
+
+class CreditInvoice(metaclass=PoolMeta):
+    __name__ = 'account.invoice.credit'
+
+    @property
+    def _credit_options(self):
+        options = super()._credit_options
+        if not self.start.with_agent:
+            options['agent'] = None
+        return options
